!!* General purpose wrapper for MD integrators.
!!* @note Currently only velocity Verlet is wrapped.
module MDIntegrator
#include "assert.h"
#include "allocate.h"  
  use Accuracy
  use VelocityVerlet
  !use VelocityOmelyan
  implicit none
  private

  public :: OMDIntegrator
  public :: create, destroy, next

  !!* Data for the MD integrator.
  type OMDIntegrator
    private
    integer :: integrator
    type(OVelocityVerlet), pointer :: pVelocityVerlet
    !type(OVelocityOmelyan), pointer :: pVelocityOmelyan
  end type OMDIntegrator

  interface create
    module procedure MDIntegrator_create_VVerlet
    !module procedure MDIntegrator_create_VOmelyan
  end interface

  interface destroy
    module procedure MDIntegrator_destroy
  end interface

  interface next
    module procedure MDIntegrator_next
  end interface
  
  !! Type of the integrator
  integer, parameter :: velocityVerlet_ = 1
!  integer, parameter :: velocityOmelyan_ = 2

contains

  !!* Create integrator wrapper for velocity Verlet.
  !!* @param self Integrator wrapper instance on exit.
  !!* @param pIntegrator Velocity Verlet integrator.
  subroutine MDIntegrator_create_VVerlet(self, pIntegrator)
    type(OMDIntegrator), pointer :: self
    type(OVelocityVerlet), pointer :: pIntegrator

    INITALLOCATE_P(self)
    self%integrator = velocityVerlet_
    self%pVelocityVerlet => pIntegrator
    
  end subroutine MDIntegrator_create_VVerlet


  
  !subroutine MDIntegrator_create_VOmelyan(self, pIntegrator)
  !  type(OMDIntegrator), pointer :: self
  !  type(OVelocityOmelyan), pointer :: pIntegrator
  !
  !  INITALLOCATE_P(self)
  !  self%integrator = velocityOmelyan_
  !  self%pVelocityOmelyan => pIntegrator
  !  
  !end subroutine MDIntegrator_create_VOmelyan


  
  !!* Destroys MD integrator.
  !!* @param self Integrator wrapper.
  subroutine MDIntegrator_destroy(self)
    type(OMDIntegrator), pointer :: self

    if (.not. associated(self)) then
      return
    end if
    select case (self%integrator)
    case (velocityVerlet_)
      call destroy(self%pVelocityVerlet)
    !case (velocityOmelyan_)
    !  call destroy(self%velocityOmelyan)
    end select
    DEALLOCATE_P(self)
    
  end subroutine MDIntegrator_destroy

  

  !!* Delivers the next velocities
  !!* @param self Integrator wrapper instance on exit.
  !!* @param accel Accelerations.
  !!* @param newCoords Updated coordinates.
  !!* @param newVelocity Updated velocities.
  subroutine MDIntegrator_next(self, accel, newCoord, newVelocity)
    type(OMDIntegrator), pointer :: self
    real(dp), intent(in) :: accel(:,:)
    real(dp), intent(out) :: newCoord(:,:)
    real(dp), intent(out) :: newVelocity(:,:)

    ASSERT(associated(self))
    
    select case (self%integrator)
    case (velocityVerlet_)
      call next(self%pVelocityVerlet, accel, newCoord, newVelocity)
    !case (velocityOmelyan_)
    !  call next(self%pVelocityOmelyan, accel, newCoord, newVelocity)
    end select
    
  end subroutine MDIntegrator_next
  
end module MDIntegrator
