!!* Global variables and initialization for the main program
!!* @todo Assignment (copy) operator for TNeighbors!!!
module initprogram
# include "assert.h"
# include "allocate.h"  
  use inputdata_
  use constants, only: pi
  use periodic
  use accuracy
  use intrinsicprinter
  use short_gamma
  use coulomb
  use message

  use Mixer
  use SimpleMixer
  use AndersonMixer
  use BroydenMixer
  use DIISMixer

  use GeoOpt
  use ConjugateGradient
  use SteepestDescent
  
  use Ranlux
  use MDCommon
  use MDintegrator
  use VelocityVerlet
  use Thermostat
  use DummyThermostat
  use AndersenThermostat
  use BerendsenThermostat
  use TempProfile

  use rcm
  use LapackRoutines
  use SimpleAlgebra

  use SCC
  use SCC_init
  use SlakoCont
  use RepCont
  use SlaterKirkwood

  use Spin, only: Spin_getOrbitalEquiv
  use OrbitalEquiv
  use CommonTypes
  
  use Fifo
  implicit none

  !! Program technical constants
  integer, parameter :: maxL = 3         !* maximal ang. momentum (s = 0, ...)


  !!* Name of the orbitals
  character(len=1), parameter :: orbitalName(maxL + 1) = &
      &(/ "S", "P", "D", "F" /)



  character(*), parameter :: fChargeIn = "charges.bin"
  character(*), parameter :: fStopDriver = "stop_driver"
  character(*), parameter :: fStopSCC = "stop_scc"

  logical               :: tSCC            !* Is the calculation SCC?
  logical               :: tDFTBU          !* is this a DFTB+U calculation?
  integer,  parameter   :: nCutoff = 1     !* Nr. of different cutoffs
  
  integer               :: nAtom           !* nr. of atoms
  integer               :: nAllAtom        !* nr. of all (image and orig) atoms
  integer,  pointer     :: Img2CentCell(:) !* nr. of original atom in centre
  integer               :: nType           !* nr of different types (nAtom)

  type(TOrbitals), pointer :: orb
  
  integer               :: nOrb            !* nr. of orbitals in the system
  integer               :: nAllOrb         !* nr. of orbitals for all atoms
  integer,  pointer     :: specie(:)       !* types of the atoms (nAllAtom)
  integer,  allocatable :: specie0(:)      !* type of the atoms (nAtom)
  real(dp), pointer     :: coord(:,:)      !* Coords of the atoms (3, nAllAtom)
  real(dp), allocatable :: coord0(:,:)     !* Initial coordinates (3, nAtom)
  logical               :: tPeriodic       !* if calculation is periodic
  logical               :: tFracCoord      !* are atomic coordinates fractional?
  real(dp)              :: sccTol          !* Tollerance for SCC cycle
  real(dp), allocatable :: latVec(:,:)     !* lattice vectors as columns (3, 3)
  real(dp), allocatable :: recVec(:,:)     !* reciprocal vecs as columns (3, 3)
  real(dp), allocatable :: recVec2p(:,:)   !* reciprocal vectors in 2pi units
  real(dp)              :: volume          !* cell volume
  real(dp)              :: recVolume       !* reciprocal cell volume
  real(dp),  pointer    :: cellVec(:,:)    !* translation vecs for interacting
                                           !* image cells (3, nImgCell + 1)
  real(dp), pointer     :: rCellVec(:,:)   !* cell vectors in absolute units
  integer,  pointer     :: iCellVec(:)     !* index in cellVec for each atom

  !!* ADT for neighbor parameters
  type(TNeighborList), save :: neighborList
  integer,  allocatable :: nNeighbor(:)    !* nr. of neighbors for SK + rep
  integer,  pointer     :: iPair(:,:)      !* H/S indexing array
      
  integer,  allocatable :: iAtomStart(:)   !* atom start pos for squared H/S

  real(dp), allocatable :: hubbU(:,:)      !* Hubbard Us (orbital, atom)
  real(dp), allocatable :: atomEigVal(:,:) !* self energy (orbital, atom)
  real(dp), allocatable :: referenceN0(:,:)!* reference n_0 charges for each
                                           !* atom  
  real(dp), allocatable :: mass(:)         !* list of atomic masses

  type(OSlakoCont), pointer :: skHamCont
  type(OSlakoCont), pointer :: skOverCont
  type(ORepCont), pointer :: pRepCont
  real(dp) :: skCutoff
  real(dp) :: skRepCutoff

  real(dp)              :: mCutoff        !* longest pair interaction
  
  real(dp), pointer     :: ham(:,:)       !* Hamiltonian
  complex(dp), pointer  :: ham_nc(:,:)    !* noncolinear Hamiltonian
  real(dp), pointer     :: h0(:)          !* Non-SCC part
  real(dp), pointer     :: h1(:)          !* SCC part
  real(dp), pointer     :: hspin(:)       !* spin part
  complex(dp), pointer  :: hspin_nc(:,:)  !* noncolinear spin part
  real(dp), pointer     :: over(:)        !* Overlap
  complex(dp), pointer  :: over_nc(:,:)   !* noncolinear Overlap


  integer               :: nKPoint        !* nr. of K-points
  real(dp), allocatable :: kPoint(:,:)    !* K-points
  real(dp), allocatable :: KWeight(:)     !* weight of the K-Points

  logical               :: tRealHS        !* H and S are real

  real(dp), allocatable :: nEl(:)         !* nr. of electrons
  real(dp)              :: nEl0           !* Nr. of all electrons if neutral
  real(dp), allocatable :: W(:,:,:)       !* Spin W's	  !'
  
  integer               :: nDFTBUfunc     !* Choice of orbital functional
  real(dp), allocatable :: UJ(:,:)        !* list of U-J for DFTB+U
  integer, allocatable  :: nU(:)          !* number of U-J for DFTB+U per species
  integer, allocatable  :: nLinU(:,:)     !* number of angular momenta in a
                                          !* given DFTB+U block
  integer, allocatable  :: iLVals(:,:,:)  !* list of angular momenta in a block 
  
  real(dp) :: tempElec                    !* electron temperature
  logical :: tFillKSep                    !* If K points should filled separat.
  logical :: tSetFillingTemp              !* Filling temp updated by MD.
  integer  :: iDistribFn = 0              !* Choice of electron distribution
                                          !* function, defaults to Fermi
  real(dp) :: tempAtom                    !* atomic temperature
  real(dp) :: deltaT                      !* MD stepsize
  integer :: solver                       !* eigensolver
  integer :: nSCCIter                     !* number of SCC iterations
  integer :: nSpin                        !* Number of spin components, 1
                                          !* is unpolarised, 2 is polarised, 4
                                          !* is noncolinear
  real(dp) :: almix

  logical               :: tGeoOpt          !* Geometry optimization needed?
  logical               :: tMD              !* Is this a MD calculation?
  logical               :: tMulliken        !* Do we need Mulliken charges?
  logical               :: tAtomicEnergy    !* Dow we need atom resolved E?
  logical               :: tEigenvecs       !* Print out eigenvectors?
  logical               :: tForces          !* Do we need forces?
  integer               :: nMovedAtom       !* Number of moved atoms
  integer, allocatable  :: indMovedAtom(:)  !* Index of the moved atoms
  integer               :: nMovedCoord      !* Nr. of moved coordinates
  integer               :: nGeoSteps        !* Nr. of geo movements to do
  integer               :: nGeoConstr       !* Nr. of geometry constraints
  integer,  allocatable :: conAtom(:)       !* Index of constrained atoms
  real(dp), allocatable :: conVec(:,:)      !* Constraint vectors

  character(lc) :: geoOutFile     !* File containing end geometry
  
  logical               :: tAppendGeo       !* Append geometries in the output?
  character(mc), allocatable :: specieName(:)

  real(dp)              :: random_pool(10)   !* pool of initial random numbers
  ! for future use. See comment in code at create(pRanlux, in this routine.

  logical            :: tInitialized = .false.
  private :: tInitialized

  type(OGeoOpt), pointer :: pGeoOpt       !* General geometry optimizer
  
  type(OMixer), pointer :: pChrgMixer    !* Charge mixer
  integer               :: iMixer        !* mixer number
  
  type(ORanlux), pointer :: pRanlux !* Random number generator

  type(OMDCommon), pointer :: pMDFrame  !* MD Framework
  type(OMDIntegrator), pointer :: pMDIntegrator !* MD integrator
  type(OTempProfile), pointer :: pTempProfile
  
      


  !! Charge related variables
  real(dp), allocatable    :: q0(:, :, :)
  real(dp), allocatable    :: qInput(:, :, :)
  real(dp), allocatable    :: qOutput(:, :, :)
  real(dp), allocatable    :: qInpRed(:), qOutRed(:), qDiffRed(:)
  integer, allocatable     :: iEqOrbitals(:,:,:)  !* Orbital equiv. relations
  integer :: nIneqOrb !* nr. of inequivalent orbitals

  !! External charges
  integer :: nExtChrg   !* Nr. of external charges
  logical :: tExtChrg   !* If external charges must be considered

  !! Other stuff
  logical :: tReadChrg    !* If initial charges/dens mtx. from external file.
  logical :: tWriteTagged !* produce tagged output?
  logical :: tWriteDetailedXML !* Produce detailed.xml
  logical :: tWriteResultsTag !* Produce detailed.tag
  logical :: tWriteDetailedOut !* Produce detailed.out
  logical :: tWriteBandDat !* Produce band.dat
  logical :: tWriteHS, tWriteRealHS  !* Should HS (square and real) be printed?
  logical :: tMinMemory, tStoreEigvecs

  !! Damped interactions
  logical, allocatable :: tDampedShort(:)


  integer :: runId !* Program run id

  integer :: restartFreq    !* Frequency for saving restart info

  logical :: tDispersion  !* If dispersion should be calculated
  type(OSlaterKirkwood), pointer :: pDispersion

  type(OFifoRealR2), allocatable :: storeEigvecsReal(:)
  type(OFifoCplxR2), allocatable :: storeEigvecsCplx(:)


  integer, parameter :: nInitNeighbor = 40  !* First guess for nr. of neighbors.
  private :: nInitNeighbor

  
contains

  
  !!* Initializes the variables in the module based on the parsed input
  !!* @param input Holds the parsed input data.
  subroutine initProgramVariables(input)
    type(inputData), intent(in) :: input

    !! Mixer related local variables
!    integer :: iMixer
    integer  :: nGeneration
    real(dp) :: mixParam
    type(OSimpleMixer), pointer :: pSimpleMixer
    type(OAndersonMixer), pointer :: pAndersonMixer
    type(OBroydenMixer), pointer :: pBroydenMixer
    type(ODIISMixer), pointer :: pDIISMixer

    !! Geometry optimizer related local variables
    type(OConjGrad), pointer :: pConjGrad    !* Conjugate gradient driver
    type(OSteepDesc), pointer :: pSteepDesc  !* Steepest descent driver

    !! MD related local variables
    type(OThermostat), pointer :: pThermostat
    type(ODummyThermostat), pointer :: pDummyTherm
    type(OAndersenThermostat), pointer :: pAndersenTherm
    type(OBerendsenThermostat), pointer :: pBerendsenTherm
    type(OVelocityVerlet), pointer :: pVelocityVerlet
    
    integer :: ii, jj, kk, iS
    character(lc) :: strTmp, strTmp2
    logical :: tFirst
    integer  :: iSeed, timeValues(8) 
    real(dp) :: rTmp



    logical :: tExist ! Flag if some files do exist or not

    !! Orbital equivalency for SCC and Spin
    integer, allocatable :: iEqOrbSCC(:,:,:), iEqOrbSpin(:,:,:)

    ASSERT(input%tInitialized)
    
    !! Basic variables
    tSCC = input%ctrl%tScc
    tDFTBU = input%ctrl%tDFTBU
    if (input%ctrl%tSpin) then
      nSpin = 2              
    else
      nSpin = 1
    end if
    ! nSpin = 4

    sccTol = input%ctrl%sccTol
    nAtom = size(input%geom%coords, dim = 2)
    nType = maxval(input%geom%species)
    tPeriodic = input%geom%tPeriodic
    tFracCoord = input%geom%tFracCoord
    solver = input%ctrl%iSolver
    if (tSCC) then
      nSCCIter = input%ctrl%maxIter
    else
      nSCCIter = 1
    end if

    if (tPeriodic) then
      ALLOCATE_(latVec, (3, 3))
      ASSERT(all(shape(input%geom%latVecs) == shape(latVec)))
      latVec(:,:) = input%geom%latVecs(:,:)
      ALLOCATE_(recVec, (3, 3))
      ALLOCATE_(recVec2p, (3, 3))
      recVec2p = latVec(:,:)
      call matinv(recVec2p)
      recVec2p = reshape(recVec2p, (/3, 3/), order=(/2, 1/))
      recVec = 2.0_dp * pi * recVec2p
      volume = determinant33(latVec)
      recVolume = determinant33(recVec)
    else
      ALLOCATE_(latVec, (0, 0))
      ALLOCATE_(recVec, (0, 0))
      ALLOCATE_(recVec2p, (0, 0))
      volume = 0.0_dp
      recVolume = 0.0_dp
    end if

    orb => input%slako%orb
        
    !! Slater-Koster tables
    skHamCont => input%slako%skHamCont
    skOverCont => input%slako%skOverCont
    pRepCont => input%slako%repCont
    
    ALLOCATE_(atomEigVal, (orb%mShell, nType))
    ASSERT(size(input%slako%skSelf, dim=1) >= orb%mShell)
    ASSERT(size(input%slako%skSelf, dim=2) == size(atomEigVal, dim=2))
    atomEigVal(:,:) = input%slako%skSelf(1:orb%mShell, :)

    ASSERT(size(input%slako%skOcc, dim=1) >= orb%mShell)
    ALLOCATE_(referenceN0,(orb%mShell, nType))
    referenceN0(:,:) = input%slako%skOcc(1:orb%mShell, :)
    ASSERT(size(input%slako%mass) == nType)
    ALLOCATE_(mass,(nType))
    mass(:) = input%slako%mass(:)
    ! Spin W's	!'
    if (nSpin > 1) then
      ALLOCATE_(W,(orb%mShell,orb%mShell,nType))
      W(:,:,:) = 0.0_dp
      do ii=1,nType
        do jj=1, orb%nShell(ii)
          do kk=1, orb%nShell(ii)
            W(jj,kk,ii)=input%ctrl%W(jj,kk,ii)
          end do
        end do
      end do
    end if

    ! DFTB+U parameters
    if (tDFTBU) then
      nDFTBUfunc = input%ctrl%DFTBUfunc
      ALLOCATE_(nU,(nType))
      nU(:) = input%ctrl%nU(:)
      ALLOCATE_(UJ,(maxval(nU),nType))
      UJ(:,:)=0.0_dp
      do ii=1,nType
        do jj = 1, nU(ii)
          UJ(jj,ii) = input%ctrl%UJ(jj,ii)
        end do
      end do
      ALLOCATE_(nLinU,(maxval(nU),nType))
      nLinU(:,:) = 0
      do ii=1,nType
        do jj = 1, nU(ii)
          nLinU(jj,ii) = input%ctrl%nLinU(jj,ii)
        end do
      end do
      ALLOCATE_(iLVals,(maxval(nLinU),maxval(nU),nType))
      iLVals(:,:,:) = 0
      do ii=1,nType
        do jj = 1, nU(ii)
          do kk = 1, nLinU(jj,ii)
            iLVals(kk,jj,ii) = input%ctrl%iLVals(kk,jj,ii)
          end do
        end do
      end do
    end if
    
    !! Cutoffs from SlaKo and repulsive
    skCutoff = max(getCutoff(skHamCont), getCutoff(skOverCont))
    skRepCutoff = max(skCutoff, getCutoff(pRepCont))
    mCutoff = skRepCutoff
    
    !! Get specie names and output file
    geoOutFile = input%ctrl%outFile
    ALLOCATE_(specieName, (size(input%geom%specieNames)))
    specieName(:) = input%geom%specieNames(:)

    !! Initialise the SCC module
    ALLOCATE_(hubbU, (orb%mShell, nType))
    ASSERT(size(input%slako%skHubbU, dim=1) >= orb%mShell)
    ASSERT(size(input%slako%skHubbU, dim=2) == size(hubbU, dim=2))
    hubbU(:,:) = input%slako%skHubbU(1:orb%mShell, :)
    if (tSCC) then
      !! Damping of short range scc (at the moment only for H)
      ALLOCATE_(tDampedShort, (nType))
      if (input%ctrl%tDampH) then
        tDampedShort(:) = (specieName == "H" .or. specieName == "h")
      else
        tDampedShort(:) = .false.
      end if

      nExtChrg = input%ctrl%nExtChrg
      tExtChrg =  nExtChrg > 0
      if (tPeriodic) then
        if (tExtChrg) then
          call init_SCC(nAtom, nType, orb%nShell, orb%mShell, orb%mOrb, hubbU, &
              &tDampedShort, latVec, recVec, volume, input%ctrl%extChrg)
        else
          call init_SCC(nAtom, nType, orb%nShell, orb%mShell, orb%mOrb, hubbU, &
              &tDampedShort, latVec, recVec, volume)
        end if
      else
        if (tExtChrg) then
          call init_SCC(nAtom, nType, orb%nShell, orb%mShell, orb%mOrb, hubbU, &
              &tDampedShort, extCharges=input%ctrl%extChrg, blurWidths=input&
              &%ctrl%extChrgBlurWidth)
        else
          call init_SCC(nAtom, nType, orb%nShell, orb%mShell, orb%mOrb, hubbU, tDampedShort)
        end if
      end if
      mCutoff = max(mCutoff, getSCCCutoff())
    end if

    !! Dispersion
    tDispersion = associated(input%ctrl%pSlaKirk)
    if (tDispersion) then
      if (tPeriodic) then
        call create(pDispersion, input%ctrl%pSlaKirk, latVec, recVec, volume)
      else
        call create(pDispersion, input%ctrl%pSlaKirk)
      end if
      mCutoff = max(mCutoff, getRCutoff(pDispersion))
    end if

    !! Initial coordinates
    ALLOCATE_(coord0, (3, nAtom))
    ASSERT(all(shape(coord0) == shape(input%geom%coords)))
    coord0(:,:) = input%geom%coords(:,:)
    ALLOCATE_(specie0, (nAtom))
    ASSERT(all(shape(specie0) == shape(input%geom%species)))
    specie0(:) = input%geom%species(:)
    if (tPeriodic) then
      !! Make some guess for the nr. of all interacting atoms
      nAllAtom = int((real(nAtom, dp)**(1.0_dp/3.0_dp) + 3.0_dp)**3)
    else
      nAllAtom = nAtom
    end if
    INITALLOCATE_PARR(coord, (3, nAllAtom))
    INITALLOCATE_PARR(specie, (nAllAtom))
    INITALLOCATE_PARR(img2CentCell, (nAllAtom))
    INITALLOCATE_PARR(iCellVec, (nAllAtom))
    ALLOCATE_(iAtomStart, (nAtom + 1))
    call buildSquaredAtomIndex(iAtomStart, orb)

    !! Initalize images (translations)
    if (tPeriodic) then
      INIT_PARR(cellVec)
      INIT_PARR(rCellVec)
      call getCellTranslations(cellVec, rCellVec, latVec, recVec2p, mCutoff)
    else
      INITALLOCATE_PARR(cellVec, (3, 1))
      INITALLOCATE_PARR(rCellVec, (3, 1))
      cellVec(:, 1) = (/ 0.0_dp, 0.0_dp, 0.0_dp /)
      rCellVec(:, 1) = (/ 0.0_dp, 0.0_dp, 0.0_dp /)
    end if

    !! Initialize neighborlist.
    call init(neighborList, nAtom, nInitNeighbor)
    ALLOCATE_(nNeighbor, (nAtom))
    
    !! Intialize Hamilton and overlap
    if (tSCC) then
      INITALLOCATE_PARR(h0, (0))
      INITALLOCATE_PARR(h1, (0))
    end if
    select case(nSpin)
    case(1)
      INITALLOCATE_PARR(ham, (0, nSpin))
      INITALLOCATE_PARR(over, (0))
    case(2)
      INITALLOCATE_PARR(ham, (0, nSpin))
      INITALLOCATE_PARR(over, (0))
      INITALLOCATE_PARR(hspin, (0))
    case(4)
      INITALLOCATE_PARR(ham_nc, (0, nSpin))
      INITALLOCATE_PARR(over_nc, (0, nSpin))
      INITALLOCATE_PARR(over, (0))
      INITALLOCATE_PARR(hspin_nc, (0, nSpin))      
    end select      
    INITALLOCATE_PARR(iPair, (0, nAtom))
    
    !! Brillouin zone sampling
    if (tPeriodic) then
      nKPoint = input%ctrl%nKPoint
      ALLOCATE_(kPoint, (3, nKPoint))
      ALLOCATE_(kWeight, (nKPoint))
      ASSERT(all(shape(kPoint) == shape(input%ctrl%KPoint)))
      ASSERT(all(shape(kWeight) == shape(input%ctrl%kWeight)))
      kPoint(:,:) = input%ctrl%KPoint(:,:)
      if (sum(input%ctrl%kWeight(:)) < epsilon(1.0_dp)) then
        call error("Sum of k-point weights should be greater than zero!")
      end if
      kWeight(:) = input%ctrl%kWeight(:) / sum(input%ctrl%kWeight(:))
    else
      nKPoint = 1
      ALLOCATE_(kPoint, (3, nKPoint))
      ALLOCATE_(kWeight, (nKpoint))
      kPoint(:,1) = 0.0_dp
      kWeight(1) = 1.0_dp
    end if

    if ((.not. tPeriodic) .or. (nKPoint == 1 .and. &
        &all(kPoint(:, 1) == (/ 0.0_dp, 0.0_dp, 0.0_dp /)))) then
      tRealHS = .true.
    else
      tRealHS = .false.
    end if

    !! Other usefull quantities
    nOrb = sum(orb%nOrbAtom(1:nAtom))

    if (nSpin < 3) then
      ALLOCATE_(nEl,(nSpin))
    else
      ALLOCATE_(nEl,(1))
    end if
    nEl0 = 0.0_dp
    do ii = 1, nAtom
      nEl0 = nEl0 + sum(input%slako%skOcc(1:orb%nShell(specie0(ii)), &
          & specie0(ii)))
    end do
    nEl(:) = 0.0_dp
    select case(nSpin)
    case(1)
      nEl(1) = 0.5_dp*(nEl0 - input%ctrl%nrChrg)
    case(2)
      nEl(1) = 0.5_dp * (nEl0 - input%ctrl%nrChrg + input%ctrl%nrSpinPol)
      nEl(2) = 0.5_dp * (nEl0 - input%ctrl%nrChrg - input%ctrl%nrSpinPol)
    case default
      nEl(1) = nEl0 - input%ctrl%nrChrg
    end select    
    if (.not.all(nEl(:) >= 0.0_dp)) then
      call error("Less than 0 electrons!")
    end if
    
    if (nSpin<3) then
      if(.not.all(ceiling(nEl) <= spread(nOrb,1,nSpin))) then
        call error("More electrons than basis functions!")
      end if
    else
      if(ceiling(nEl(1)) > 2*nOrb) then
        call error("More electrons than basis functions!")
      end if
    end if

    iDistribFn = input%ctrl%iDistribFn
    tempElec = input%ctrl%tempElec
    tSetFillingTemp = input%ctrl%tSetFillingTemp
    tFillKSep = input%ctrl%tFillKSep
    tempAtom = input%ctrl%tempAtom
    deltaT = input%ctrl%deltaT
    
    !! banded matrix case, reorder to minimize bandwidth
    if (solver == 5) then 
      call reorderAtoms(coord0,specie0,neighborList%iNeighbor,nNeighbor,&
          &img2CentCell)
    end if

    !! Create equivalency relations
    if (tSCC) then
      ALLOCATE_(iEqOrbitals, (orb%mOrb, nAtom, nSpin))
      ALLOCATE_(iEqOrbSCC, (orb%mOrb, nAtom, nSpin))
      call SCC_getOrbitalEquiv(orb, specie0, iEqOrbSCC)
      select case (nSpin)
      case (1)
        iEqOrbitals(:,:,:) = iEqOrbSCC(:,:,:)
      case (2)
        ALLOCATE_(iEqOrbSpin, (orb%mOrb, nAtom, nSpin))
        call Spin_getOrbitalEquiv(orb, specie0, iEqOrbSpin)
        call OrbitalEquiv_merge(iEqOrbSCC, iEqOrbSpin, orb, iEqOrbitals)
        DEALLOCATE_(iEqOrbSpin)
      case default
        call error("Equivalency not working for non-collinear spin")
      end select
      DEALLOCATE_(iEqOrbSCC)
      nIneqOrb = maxval(iEqOrbitals)
    else
      nIneqOrb = nOrb
    end if


    !! Initialize mixer
    !! (at the moment, the mixer is not aware about any equivalency relations,
    !! the entire chrg vectors are mixed, and then the error is calculated by
    !! using the equivalency relations between the orbitals)
    if (tSCC) then
      iMixer = input%ctrl%iMixSwitch
      nGeneration = input%ctrl%iGenerations
      mixParam = input%ctrl%almix
      select case (iMixer)
      case (1)
        call create(pSimpleMixer, mixParam)
        call create(pChrgMixer, pSimpleMixer)
      case (2)
        if (input%ctrl%andersonNrDynMix > 0) then
          call create(pAndersonMixer, nGeneration, mixParam, &
              &input%ctrl%andersonInitMixing, input%ctrl%andersonDynMixParams, &
              &input%ctrl%andersonOmega0)
        else
          call create(pAndersonMixer, nGeneration, mixParam, &
              &input%ctrl%andersonInitMixing, omega0=input%ctrl%andersonOmega0)
        end if
        call create(pChrgMixer, pAndersonMixer)
      case (3)
        call create(pBroydenMixer, nSCCIter, mixParam, &
            &input%ctrl%broydenOmega0, input%ctrl%broydenMinWeight, &
            &input%ctrl%broydenMaxWeight, input%ctrl%broydenWeightFac, &
            &nGeneration)
        call create(pChrgMixer, pBroydenMixer)
      case(4)
        call create(pDIISMixer,nGeneration, mixParam, input%ctrl%tFromStart)
        call create(pChrgMixer, pDIISMixer)
      case default
        call error("Unknown charge mixer type.")
      end select
    else
      INIT_P(pChrgMixer)
    end if
    
    !! initialise in cases where atoms move
    tGeoOpt = input%ctrl%tGeoOpt
    tAppendGeo = input%ctrl%tAppendGeo
    tMD = input%ctrl%tMD
    tMulliken = input%ctrl%tMulliken
    tAtomicEnergy = input%ctrl%tAtomicEnergy
    tEigenvecs = input%ctrl%tEigenvecs
    tForces = input%ctrl%tForces   
    
    nMovedAtom = input%ctrl%nrMoved
    nMovedCoord = 3 * nMovedAtom
    nGeoSteps = input%ctrl%maxRun
    if (nMovedAtom > 0) then
      ALLOCATE_(indMovedAtom, (size(input%ctrl%indMovedAtom)))
      indMovedAtom(:) = input%ctrl%indMovedAtom(:)
    else
      ALLOCATE_(indMovedAtom, (0))
    end if

    INIT_P(pGeoOpt)
    if (nMovedAtom > 0) then
      !! Index of moved atom explicitely stored. (Could be usefull later
      !! if specifying moved atoms gets more flexible in the input.)
      if (tGeoOpt) then
        select case (input%ctrl%iGeoOpt)
        case(1)
          call create(pSteepDesc, nMovedCoord, input%ctrl%maxForce, 0.2_dp,&
              &0.5_dp * input%ctrl%deltaT**2&
              &/reshape(spread(mass(specie0(indMovedAtom(:))), 1, 3), &
              &(/nMovedCoord/)))
          call create(pGeoOpt, pSteepDesc)
        case (2)
          call create(pConjGrad, nMovedCoord, input%ctrl%maxForce, 0.2_dp)
          call create(pGeoOpt, pConjGrad)
        end select
        call reset(pGeoOpt, reshape(coord0(:, indMovedAtom), (/ nMovedCoord /)))
      end if
    else
      nMovedAtom = 0
      nGeoSteps = 0
    end if

    !! Initialize constraints
    nGeoConstr = input%ctrl%nrConstr
    if (nGeoConstr > 0) then
      ALLOCATE_(conAtom, (input%ctrl%nrConstr))
      ALLOCATE_(conVec, (3, input%ctrl%nrConstr))
      ASSERT(all(shape(conAtom) == shape(input%ctrl%conAtom)))
      ASSERT(all(shape(conVec) == shape(input%ctrl%conVec)))
      conAtom(:) = input%ctrl%conAtom(:)
      conVec(:,:) = input%ctrl%conVec(:,:)
      do ii = 1, nGeoConstr
        conVec(:,ii) = conVec(:,ii) / sqrt(sum(conVec(:,ii)**2))
      end do
    else
      ALLOCATE_(conAtom, (0))
      ALLOCATE_(conVec, (3, 0))
    end if

    !! Generate a random id for the run. Seed with system time if possible.
    call system_clock(iSeed)
    !! Try date_and_time if system_clock does not work properly.
    if (iSeed < 1) then
      call date_and_time(values=timeValues)
      if (timeValues(5) >= 0) then
        iSeed = 1000 * (60 * (60 * timeValues(5) + timeValues(6)) &
            &+ timeValues(7)) + timeValues(8)
      end if
    end if
    !! Try  default seeder if attempts to use the clock failed.
    if (iSeed < 1) then                
      call random_seed
      call random_number(rTmp)
      ! Make sure seed > 0.
      iSeed = int(real(huge(iSeed) - 1, dp) * rTmp) + 1
    end if
    call create(pRanlux, initSeed=iSeed)
    call getRandom(pRanlux, rTmp)
    runId = int(real(huge(runId) - 1, dp) * rTmp) + 1
    call destroy(pRanlux)

    !! Create random generator and pull off first 10
    !! random numbers to avoid disturbing the subsequent sequence.
    !! random_pool can then be used to seed other generators if needed,
    !! with a supply of random numbers controlled from the initial seed.
    !! If the size of random_pool is changed then reproducibility of
    !! the random numbers if initialised from a seed is lost.
    iSeed = input%ctrl%iSeed
    if (iSeed < 1) then
      iSeed = runId     ! No seed specified, use random runId
    end if
    call create(pRanlux, 3, iSeed)
    call getRandom(pRanlux,random_pool(:))
    

    !! MD stuff
    if (tMD) then
      !! Create MD framework.
      call create(pMDFrame, nMovedAtom, nAtom, input%ctrl%tMDstill)
      
      !! Create temperature profile, if thermostat is not the dummy one
      if (input%ctrl%iThermostat /= 0) then
        call create(pTempProfile, input%ctrl%tempMethods, &
            &input%ctrl%tempSteps, input%ctrl%tempValues)
      else
        INIT_P(pTempProfile)
      end if

      !! Create thermostat
      select case (input%ctrl%iThermostat)
      case (0) ! No thermostat
        call create(pDummyTherm, tempAtom, mass(specie0(indMovedAtom)), &
            &pRanlux, pMDFrame)
        call create(pThermostat, pDummyTherm)
      case (1) ! Andersen thermostat
        call create(pAndersenTherm, pRanlux, mass(specie0(indMovedAtom)), &
            &pTempProfile, input%ctrl%tRescale, input%ctrl%wvScale, pMDFrame)
        call create(pThermostat, pAndersenTherm)
      case (2) ! Berendsen thermostat
        call create(pBerendsenTherm, pRanlux, mass(specie0(indMovedAtom)), &
            &pTempProfile, input%ctrl%wvScale, pMDFrame)
        call create(pThermostat, pBerendsenTherm) 
      end select

      !! Create MD integrator
      if (input%ctrl%tReadMDVelocities) then
        call create(pVelocityVerlet, deltaT, coord0(:,indMovedAtom),&
            & pThermostat,input%ctrl%initialVelocities) 
      else
        call create(pVelocityVerlet, deltaT, coord0(:,indMovedAtom),&
            & pThermostat)
      end if
      call create(pMDIntegrator, pVelocityVerlet)
    else
      INIT_P(pMDIntegrator)
    end if
    

    !! Allocate charge arrays
    if (tMulliken) then ! automatically true if tSCC
      ALLOCATE_(q0, (orb%mOrb, nAtom, nSpin))
      q0(:,:,:) = 0.0_dp
      ALLOCATE_(qInput, (orb%mOrb, nAtom, nSpin))
      ALLOCATE_(qOutput, (orb%mOrb, nAtom, nSpin))
      qInput(:,:,:) = 0.0_dp
      qOutput(:,:,:) = 0.0_dp
    end if
    if (tSCC .and. .not. tDFTBU) then
      ALLOCATE_(qDiffRed, (maxval(iEqOrbitals)))
      ALLOCATE_(qInpRed, (maxval(iEqOrbitals)))
      ALLOCATE_(qOutRed, (maxval(iEqOrbitals)))
    end if

    !! Initialize charges (density matrix initialization postponed, since
    !! the size of the matrix is not known yet :-(
    if (tMulliken) then
      call initializeCharges(q0, referenceN0, specie0, orb)
    end if
    tReadChrg = input%ctrl%tReadChrg
    if (tSCC) then
      if (tReadChrg) then
        print *, "INIT CHARGE"
        call initializeCharges(qInput, fChargeIn, nEl, orb)
        print *, qInput
      else
        qInput(:,:,:) = q0(:,:,:)
      end if
    end if
    if (tSCC) then
      if (associated(input%ctrl%initialSpins) .and. nSpin > 1) then          
        do ii = 1, nAtom
          !! doesn't actually matter if additional spin
          !! polarization pushes charges to <0 as the initial charges aren't
          !! mixed in to later iterations            
          qInput(1:orb%nOrbAtom(ii),ii,1) = &
              & qInput(1:orb%nOrbAtom(ii),ii,1) &
              &+ 0.5_dp * input%ctrl%initialSpins(ii)/orb%nOrbAtom(ii)
          qInput(1:orb%nOrbAtom(ii),ii,2) = &
              & qInput(1:orb%nOrbAtom(ii),ii,2) &
              & - 0.5_dp * input%ctrl%initialSpins(ii)/orb%nOrbAtom(ii)
        end do
      end if
      if (.not. tDFTBU) then
        call OrbitalEquiv_reduce(qInput, iEqOrbitals, orb, qInpRed)
      end if
    end if
    
    !! Set various options
    tWriteTagged = input%ctrl%tWriteTagged
    tWriteDetailedXML = input%ctrl%tWriteDetailedXML
    tWriteResultsTag = input%ctrl%tWriteResultsTag
    tWriteDetailedOut = input%ctrl%tWriteDetailedOut
    tWriteBandDat = input%ctrl%tWriteBandDat
    tWriteHS = input%ctrl%tWriteHS
    tWriteRealHS = input%ctrl%tWriteRealHS

    !! Minimize memory usage?
    tMinMemory = input%ctrl%tMinMemory
    tStoreEigvecs = tMinMemory .and. (nKPoint > 1 .or. nSpin > 1)
    if (tStoreEigvecs) then
      if (tRealHS) then
        ALLOCATE_(storeEigvecsReal, (nSpin))
        do iS = 1, nSpin
          call init(storeEigvecsReal(iS), 0, "tmp_eigvr_")
        end do
      else
        ALLOCATE_(storeEigvecsCplx, (nSpin))
        do iS = 1, nSpin
          call init(storeEigvecsCplx(iS), 0, "tmp_eigvc_")
        end do
      end if
    end if
    
    !! Check if stopfiles already exist and quit if yes
    inquire(file=fStopSCC, exist=tExist)
    if (tExist) then
      call error("Stop file '" // fStopSCC // "' already present at startup")
    end if
    inquire(file=fStopDriver, exist=tExist)
    if (tExist) then
      call error("Stop file '" // fStopDriver // "' already present at startup")
    end if

    restartFreq = input%ctrl%restartFreq
    
    tInitialized = .true.


    if (input%ctrl%tMD) then
      select case(input%ctrl%iThermostat)
      case (0)
        write(*,"('Mode:',T30,A,/,T30,A)") 'MD without scaling of velocities',&
            & '(a.k.a. NVE ensemble)'
      case (1)
        write(*,"('Mode:',T30,A,/,T30,A)") &
            & "MD with re-selection of velocities according to temperature", &
            & "(a.k.a. NVT ensemble using Andersen thermostating)"
      case(2)
        write(*,"('Mode:',T30,A,/,T30,A)") &
            & "MD with scaling of velocities according to temperature", &
            & "(a.k.a. not NVT ensemble using Berendsen thermostating)"
      case default
        call error("Unknown thermostat mode")
      end select
    elseif (tGeoOpt) then
      if (nGeoConstr > 0) then
        strTmp = "with constraints"
      else
        strTmp = ""
      end if
      select case (input%ctrl%iGeoOpt)
      case (1) 
        write(*,"('Mode:',T30,A)")'Steepest descent' // trim(strTmp)
      case (2)
        write(*,"('Mode:',T30,A)") 'Conjugate gradient relaxation' &
            &// trim(strTmp)
      case default
        call error("Unknown optimisation mode")
      end select
    else
      write (*,"('Mode:',T30,A)") "Static calculation"
    end if

    if (tSCC) then
      write (*, "(A,':',T30,A)") "Self consistent charges", "Yes"
      write (*, "(A,':',T30,E14.6)") "SCC-tolerance", sccTol
      write (*, "(A,':',T30,I14)") "Max. scc iterations", nSCCIter
      if (tDFTBU) then
        write (*, "(A,':',T35,A)") "Orbitally dependant functional", "Yes"
        write (*, "(A,':',T30,I14)") "Orbital functional number",nDFTBUfunc
      end if
    else
      write (*, "(A,':',T30,A)") "Self consistent charges", "No"
    end if
    
    select case (nSpin)
    case(1)
      write (*,"(A,':',T30,A)") "Spin polarisation", "No"
      write (*, "(A,':',T30,F12.6,/,A,':',T30,F12.6)") "Nr. of up electrons", &
          &nEl(1), "Nr. of down electrons", nEl(1)
    case(2)
      write (*,"(A,':',T30,A)") "Spin polarisation", "Yes"
      write (*, "(A,':',T30,F12.6,/,A,':',T30,F12.6)") "Nr. of up electrons", &
          &nEl(1), "Nr. of down electrons", nEl(2)
    case(4)
      write (*,"(A,':',T30,A)") "Spin polarisation", "Yes"
      write (*,"(A,':',T30,A)") "Colinear spin", "No"
      write (*, "(A,':',T30,F12.6)") "Nr. of electrons", nEl(1)
    end select

    if (tPeriodic) then
      write (*, "(A,':',T30,A)") "Periodic boundaries", "Yes"
    else
      write (*, "(A,':',T30,A)") "Periodic boundaries", "No"
    end if

    select case (solver)
    case(1)
      write (strTmp, "(A)") "Standard"
    case(2)
      write (strTmp, "(A)") "Divide and Conquer"
    case(3)
      write (strTmp, "(A)") "Relatively robust (version 1)"
    case(4)
      write (strTmp, "(A)") "Relativel robust (version 2)"
    case(5)
      write (strTmp, "(A)") "Banded"
    end select
    write (*, "(A,':',T30,A)") "Diagonalizer", trim(strTmp)

    if (tSCC) then
      select case (iMixer)
      case(1)
        write (strTmp, "(A)") "Simple"
      case(2)
        write (strTmp, "(A)") "Anderson"
      case(3)
        write (strTmp, "(A)") "Broyden"
      case(4)
        write (strTmp, "(A)") "DIIS"
      end select
      write (*, "(A,':',T30,A,' ',A)") "Mixer", trim(strTmp), "mixer"
      write (*, "(A,':',T30,F14.6)") "Mixing parameter", mixParam
      write (*, "(A,':',T30,I14)") "Maximal SCC-cycles", nSCCIter
      select case (iMixer)
      case(2)
        write (*, "(A,':',T30,I14)") "Nr. of chrg. vectors to mix", nGeneration
      case(3)
        write (*, "(A,':',T30,I14)") "Nr. of chrg. vec. in memory", nGeneration
      case(4)
        write (*, "(A,':',T30,I14)") "Nr. of chrg. vectors to mix", nGeneration
      end select
    end if

    if (tGeoOpt) then
      write (*, "(A,':',T30,I14)") "Nr. of moved atoms", nMovedAtom
      write (*, "(A,':',T30,I14)") "Max. nr. of geometry steps", nGeoSteps
      write (*, "(A,':',T30,E14.6)") "Force tolerance", input%ctrl%maxForce
      if (input%ctrl%iGeoOpt == 1) then
        write (*, "(A,':',T30,E14.6)") "Step size", input%ctrl%deltaT
      end if
    end if
    

    tFirst = .true.
    if (nGeoConstr > 0) then
      do ii = 1, nAtom
        do jj = 1, nGeoConstr
          if (conAtom(jj) == ii) then
            if (tFirst) then
              write(strTmp, "(A,':')") "Geometry constraints"
              tFirst = .false.
            else
              write(strTmp, "(A)") ""
            end if
            write(*, "(A,T30,'At',I4,': ',3F10.6)") trim(strTmp), &
                &ii, (conVec(kk,jj), kk=1,3)
          end if
        end do
      end do
    end if

    if (.not.input%ctrl%tSetFillingTemp) then
      write (*, "(A,':',T30,E14.6)") "Electronic temperature", tempElec
    end if
    if (tMD) then
      write (*, "(A,':',T30,E14.6)") "Time step", deltaT
      if (input%ctrl%iThermostat == 0 .and. .not.input%ctrl%tReadMDVelocities)&
          & then 
        write (*, "(A,':',T30,E14.6)") "Temperature", tempAtom
      end if
      write (*, "(A,':',T30,I14)") "Random seed", iSeed
      if (input%ctrl%tRescale) then
        write (*, "(A,':',T30,E14.6)") "Rescaling probability", &
            &input%ctrl%wvScale
      end if
    end if

    if (tSCC) then
      if (input%ctrl%tReadChrg) then
        write (strTmp, "(A,A,A)") "Read in from '", trim(fChargeIn), "'"
      else
        write (strTmp, "(A,E11.3,A)") "Set automatically (system chrg: ", &
            &input%ctrl%nrChrg, ")"
      end if
      write (*, "(A,':',T30,A)") "Initial charges", trim(strTmp)
    end if
    
    do ii = 1, nType
      if (ii == 1) then
        write (strTmp, "(A,':')") "Included shells"
      else
        write (strTmp, "(A)") ""
      end if
      do jj = 1, orb%nShell(ii)
        if (jj == 1) then
          strTmp2 = trim(orbitalName(orb%angShell(jj, ii) + 1))
        else
          strTmp2 = trim(strTmp2) // ", " &
              &// trim(orbitalName(orb%angShell(jj, ii) + 1))
        end if
      end do
      write (*, "(A,T30,A2,':  ',A)") trim(strTmp), trim(specieName(ii)), &
          &trim(strTmp2)
    end do

    if (tPeriodic) then
      do ii = 1, nKPoint
        if (ii == 1) then
          write(strTmp, "(A,':')") "K-points and weights"
        else
          write(strTmp, "(A)") ""
        end if
        write(*,"(A,T30,I4,':',3F10.6,3X,F10.6)") trim(strTmp), ii, &
            & (kPoint(jj, ii), jj=1, 3), kWeight(ii)
      end do
    end if

    tFirst = .true.
    if (nSpin > 1) then
      do ii = 1, nType
        do jj = 1, orb%nShell(ii)
          do kk = 1, orb%nShell(ii)
            if (tFirst) then
              write(strTmp, "(A,':')") "Spin coupling constants"
              tFirst = .false.
            else
              write(strTmp, "(A)") ""
            end if
            write(*, "(A,T30,A2,2X,I1,'(',A1,')-',I1,'(',A1,'): ',E14.6)") &
                &trim(strTmp), specieName(ii), &
                jj, orbitalName(orb%angShell(jj, ii)+1), &
                &kk, orbitalName(orb%angShell(kk, ii)+1), &
                &W(kk, jj, ii)
          end do
        end do
      end do
    end if

    write (*, "(A,':')") "Extra options"
    if (tMulliken .and. .not. tSCC) then
      write (*, "(T30,A)") "Mulliken analysis"
    end if
    if (tForces .and. .not. (tMD .or. tGeoOpt)) then
      write (*, "(T30,A)") "Force calculation"
    end if
    if (tEigenvecs) then
      write (*, "(T30,A)") "Eigenvector printing"
    end if
    if (tExtChrg) then
      write (*, "(T30,A)") "External charges specified"
    end if

    
    tInitialized = .true.
    
  end subroutine initProgramVariables

  



  !!* Destroys the program variables
  subroutine destroyProgramVariables

    integer :: iS

    ASSERT(tInitialized)

    tInitialized = .false.

    DEALLOCATE_(latVec)
    DEALLOCATE_(recVec)
    DEALLOCATE_(recVec2p)
    DEALLOCATE_(hubbU)
    DEALLOCATE_(atomEigVal)
    DEALLOCATE_(referenceN0)
    DEALLOCATE_(mass)
    DEALLOCATE_(coord0)
    DEALLOCATE_PARR(Img2CentCell)
    DEALLOCATE_PARR(cellVec)
    DEALLOCATE_PARR(rCellVec)
    DEALLOCATE_PARR(iCellVec)
    DEALLOCATE_PARR(coord)
    DEALLOCATE_(coord0)
    DEALLOCATE_PARR(specie)
    DEALLOCATE_(specie0)
    DEALLOCATE_PARR(iPair)
    DEALLOCATE_(iAtomStart)
    select case (nSpin)
    case(1)
      DEALLOCATE_PARR(ham)
      DEALLOCATE_PARR(over)
    case (2)
      DEALLOCATE_PARR(ham)
      DEALLOCATE_PARR(over)
      DEALLOCATE_PARR(hspin)  
      DEALLOCATE_(W)
    case(4)
      DEALLOCATE_PARR(ham_nc)
      DEALLOCATE_PARR(over_nc)
      DEALLOCATE_PARR(over)
      DEALLOCATE_PARR(hspin_nc)
      DEALLOCATE_(W)
    end select
    if (tSCC) then
      DEALLOCATE_PARR(h0)
      DEALLOCATE_PARR(h1)
      if (tDFTBU) then
        DEALLOCATE_(UJ)
        DEALLOCATE_(nU)
        DEALLOCATE_(nLinU)
        DEALLOCATE_(iLVals)
      end if
    end if
    DEALLOCATE_(kPoint)
    DEALLOCATE_(kWeight)
    DEALLOCATE_(nEl)
    DEALLOCATE_(specieName)
    DEALLOCATE_(indMovedAtom)
    DEALLOCATE_(conAtom)
    DEALLOCATE_(conVec)

    DEALLOCATE_(qOutput)
    DEALLOCATE_(qInput)
    DEALLOCATE_(q0)
    DEALLOCATE_(qDiffRed)
    DEALLOCATE_(qInpRed)
    DEALLOCATE_(qOutRed)
    DEALLOCATE_(iEqOrbitals)

    call destroy(neighborList)
    call destroy(pChrgMixer)
    if (tGeoOpt) then
      call destroy(pGeoOpt)
    end if
    call destroy(pRanlux)
    if (tMD) then
      call destroy(pMDFrame)
      call destroy(pMDIntegrator)
    end if
    
    if (tStoreEigvecs) then
      if (tRealHS) then
        do iS = 1, nSpin
          call destruct(storeEigvecsReal(iS))
        end do
        DEALLOCATE_(storeEigvecsReal)
      else
        do iS = 1, nSpin
          call destruct(storeEigvecsCplx(iS))
        end do
        DEALLOCATE_(storeEigvecsCplx)
      end if
    end if


  end subroutine destroyProgramVariables
  
end module initprogram
